/*
 * Segment 5 Intelligente Analyseverfahren, SW-Einheit 5.2 Straßensubsegmentanalyse
 * Copyright (C) 2007-2018 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.subsegana.subsegana.application;

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.iav.fuzzylib.fuzzylib.*;
import de.bsvrz.iav.subsegana.subsegana.domain.shared.*;
import de.bsvrz.iav.subsegana.subsegana.domain.umfelddaten.*;
import de.bsvrz.iav.subsegana.subsegana.domain.verkehrsdaten.*;
import de.bsvrz.iav.subsegana.subsegana.portals.*;
import de.bsvrz.sys.funclib.application.*;
import de.bsvrz.sys.funclib.commandLineArgs.*;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList.*;
import de.bsvrz.sys.funclib.debug.*;

import java.time.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.stream.*;

/**
 * Main-Klasse zum Setzen von Standardparametern für die Straßensubsegmentanalyse.
 * <p>
 * Die folgende Abbildung zeigt den Datenfluss des Standardparametersetzers.
 * <p>
 * <img src="doc-files/standardparametersetzer-datenfluss.png" alt="Datenfluss">
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
@SuppressWarnings({"ClassDataAbstractionCoupling", "ClassFanOutComplexity"})
public final class Standardparametersetzer implements StandardApplication {

    private Debug log;

    private Collection<String> argSystemobjekte;

    private FuzzySetsSender fuzzySetsSender;
    private RegelbasisSender regelbasisSender;

    private Standardparametersetzer() {
        // Konstruktor verstrecken
    }

    public static void main(final String[] args) {
        StandardApplicationRunner.run(new Standardparametersetzer(), args);
    }

    @Override
    public void parseArguments(final ArgumentList argumentList) {
        log = Debug.getLogger();
        argSystemobjekte = bestimmeZuVerarbeitendeSystemobjekte(argumentList);
    }

    private static Collection<String> bestimmeZuVerarbeitendeSystemobjekte(ArgumentList argumentList) {
        Argument arg = argumentList.fetchArgument("-objekte=");
        if (arg.asString().trim().isEmpty()) {
            return Collections.emptyList();
        }

        String[] pids = arg.asString().split(",");
        return Arrays.stream(pids)
                .map(String::trim)
                .distinct()
                .collect(Collectors.toList());
    }

    @Override
    public void initialize(final ClientDavInterface connection) throws Exception {
        DataModel model = connection.getDataModel();
        ObjektFactory objektFactory = new ObjektFactory(model);
        Collection<Systemobjekt> objekte = objektFactory.bestimmeObjekte(argSystemobjekte);

        fuzzySetsSender = new FuzzySetsSender(connection);
        fuzzySetsSender.anmelden(objekte);
        log.info("Fuzzy-Sets-Sender bereit.");

        regelbasisSender = new RegelbasisSender(connection);
        regelbasisSender.anmelden(objekte);
        log.info("Regelbasissender bereit.");

        sendeParameter(objekte);
    }

    private void sendeParameter(Collection<Systemobjekt> objekte) throws InterruptedException {
        LocalDateTime zeitstempel = LocalDateTime.now();

        Collection<Datensatz<?, LinguistischeVariable>> fuzzySets = objekte.stream()
                .flatMap(obj -> bestimmeFuzzySets(obj, zeitstempel))
                .collect(Collectors.toList());
        fuzzySetsSender.sende(fuzzySets);

        Collection<Datensatz<?, Regelbasis>> regelbasen = objekte.stream()
                .map(obj -> bestimmeRegelbasis(obj, zeitstempel))
                .filter(d -> d.getDatum().nonNull())
                .collect(Collectors.toList());
        regelbasisSender.sende(regelbasen);

        log.info("Standardparameter gesendet");
        TimeUnit.SECONDS.sleep(20);
        System.exit(0);
    }

    private static Stream<Datensatz<?, LinguistischeVariable>> bestimmeFuzzySets(Systemobjekt objekt, LocalDateTime zeitstempel) {
        Collection<Datensatz<?, LinguistischeVariable>> fuzzySets = new ArrayList<>();
        if (objekt instanceof Stoerfallindikator) {
            fuzzySets.addAll(bestimmeFuzzySets(zeitstempel, (Stoerfallindikator) objekt));
        } else if (objekt instanceof Umfelddatenmessstelle) {
            fuzzySets.addAll(bestimmeFuzzySets(zeitstempel, (Umfelddatenmessstelle) objekt));
        } else if (objekt instanceof Umfelddatensensor) {
            fuzzySets.addAll(bestimmeFuzzySets(zeitstempel, (Umfelddatensensor) objekt));
        }
        return fuzzySets.stream();
    }

    @SuppressWarnings("CyclomaticComplexity")
    private static Collection<Datensatz<?, LinguistischeVariable>> bestimmeFuzzySets(LocalDateTime zeitstempel, Stoerfallindikator stoerfallindikator) {
        Collection<Datensatz<?, LinguistischeVariable>> fuzzySets = new ArrayList<>();
        fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.anteilLkw()));
        fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.verkehrsunruhe()));

        if (stoerfallindikator.getAnzahlFahrstreifen() == 2 && stoerfallindikator.getTyp() == Segmenttyp.FREIE_STRECKE) {
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.geschwindigkeitSzenario1()));
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.bemessungsdichteSzenario1()));
        } else if (stoerfallindikator.getAnzahlFahrstreifen() == 3 && stoerfallindikator.getTyp() == Segmenttyp.FREIE_STRECKE) {
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.geschwindigkeitSzenario2()));
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.bemessungsdichteSzenario2()));
        } else if (stoerfallindikator.getAnzahlFahrstreifen() >= 4 && stoerfallindikator.getTyp() == Segmenttyp.FREIE_STRECKE) {
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.geschwindigkeitSzenario3()));
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.bemessungsdichteSzenario3()));
        } else if (stoerfallindikator.getAnzahlFahrstreifen() == 1 && stoerfallindikator.getTyp() == Segmenttyp.VERFLECHTUNGSBEREICH) {
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.geschwindigkeitSzenario4()));
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.bemessungsdichteSzenario4()));
        } else if (stoerfallindikator.getAnzahlFahrstreifen() == 2 && stoerfallindikator.getTyp() == Segmenttyp.VERFLECHTUNGSBEREICH) {
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.geschwindigkeitSzenario5()));
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.bemessungsdichteSzenario5()));
        } else if (stoerfallindikator.getAnzahlFahrstreifen() >= 3 && stoerfallindikator.getTyp() == Segmenttyp.VERFLECHTUNGSBEREICH) {
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.geschwindigkeitSzenario6()));
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.bemessungsdichteSzenario6()));
        } else {
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.geschwindigkeitSzenario1()));
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.bemessungsdichteSzenario1()));
        }

        if (stoerfallindikator.getAnzahlFahrstreifen() == 2) {
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.verkehrsstaerkeSzenario1()));
        } else if (stoerfallindikator.getAnzahlFahrstreifen() == 3) {
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.verkehrsstaerkeSzenario2()));
        } else if (stoerfallindikator.getAnzahlFahrstreifen() >= 4) {
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.verkehrsstaerkeSzenario3()));
        } else {
            fuzzySets.add(new Datensatz<>(zeitstempel, stoerfallindikator, VerkehrsdatenStandardparameterFactory.verkehrsstaerkeSzenario1()));
        }

        return fuzzySets;
    }

    private static Collection<? extends Datensatz<?, LinguistischeVariable>> bestimmeFuzzySets(LocalDateTime zeitstempel, Umfelddatenmessstelle ufdms) {
        Collection<Datensatz<?, LinguistischeVariable>> fuzzySets = new ArrayList<>();
        fuzzySets.add(new Datensatz<>(zeitstempel, ufdms, UmfelddatenStandardparameterFactory.glaette()));
        return fuzzySets;
    }

    private static Collection<? extends Datensatz<?, LinguistischeVariable>> bestimmeFuzzySets(LocalDateTime zeitstempel, Umfelddatensensor ufds) {
        Collection<Datensatz<?, LinguistischeVariable>> fuzzySets = new ArrayList<>();
        if (ufds instanceof Helligkeitssensor) {
            fuzzySets.add(new Datensatz<>(zeitstempel, ufds, UmfelddatenStandardparameterFactory.helligkeit()));
        } else if (ufds instanceof Niederschlagsintensitaetssensor) {
            fuzzySets.add(new Datensatz<>(zeitstempel, ufds, UmfelddatenStandardparameterFactory.niederschlagsintensitaet()));
        } else if (ufds instanceof Sichtweitesensor) {
            fuzzySets.add(new Datensatz<>(zeitstempel, ufds, UmfelddatenStandardparameterFactory.sichtweite()));
        } else if (ufds instanceof Windgeschwindigkeitssensor) {
            fuzzySets.add(new Datensatz<>(zeitstempel, ufds, UmfelddatenStandardparameterFactory.windgeschwindigkeit()));
        } else if (ufds instanceof Windrichtungssensor) {
            fuzzySets.add(new Datensatz<>(zeitstempel, ufds, UmfelddatenStandardparameterFactory.windrichtung()));
        }
        return fuzzySets;
    }

    private static Datensatz<Systemobjekt, Regelbasis> bestimmeRegelbasis(Systemobjekt objekt, LocalDateTime zeitstempel) {
        Regelbasis regelbasis = Regelbasis.NULL;
        if (objekt instanceof Messquerschnitt || objekt instanceof Strassenteilsegment) {
            regelbasis = VerkehrsdatenStandardparameterFactory.regelbasis();
        } else if (objekt instanceof Umfelddatenmessstelle) {
            regelbasis = UmfelddatenStandardparameterFactory.regelbasis();
        }
        return new Datensatz<>(zeitstempel, objekt, regelbasis);
    }

}
