/*
 * Segment 5 Intelligente Analyseverfahren, SW-Einheit 5.2 Straßensubsegmentanalyse
 * Copyright (C) 2007-2018 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.subsegana.subsegana.domain;

import de.bsvrz.iav.fuzzylib.fuzzylib.*;
import de.bsvrz.iav.subsegana.subsegana.domain.shared.*;
import de.bsvrz.iav.subsegana.subsegana.domain.umfelddaten.*;
import de.bsvrz.iav.subsegana.subsegana.shared.*;

import java.time.*;
import java.util.*;
import java.util.stream.*;

/**
 * Leitet neue Fuzzy-Werte mit Hilfe einer Regelbasis ab.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class Inferenz {

    private final Event<Collection<Datensatz<?, FuzzyVariable>>> fuzzyWerteAbgeleitet = new Event<>();

    private Map<Systemobjekt, Regelbasis> regelbasen = new LinkedHashMap<>();

    private final FuzzyWerteCache fuzzyWerteCache;

    public Inferenz(int maxZyklen) {
        fuzzyWerteCache = new FuzzyWerteCache(maxZyklen);
    }

    public void regelbasisAktualisiert(Datensatz<?, Regelbasis> parameter) {
        if (parameter.getDatum().nonNull()) {
            regelbasen.put(parameter.getSystemobjekt(), parameter.getDatum());
        } else {
            regelbasen.remove(parameter.getSystemobjekt());
        }
    }

    public void werteRegelnAus(Collection<Datensatz<?, FuzzyVariable>> fuzzyWerte) {
        fuzzyWerteCache.cacheFuzzyWerte(fuzzyWerte);

        Datensatz<?, FuzzyVariable> ersterDatensatz = fuzzyWerte.iterator().next();
        Systemobjekt objekt = bestimmeObjekt(ersterDatensatz);
        Regelbasis regelbasis = regelbasen.get(objekt);
        if (regelbasis == null) {
            return;
        }

        Map<String, ?> kontext = erzeugeKontext(objekt);
        Collection<FuzzyVariable> konklusionen = regelbasis.interpretiere(kontext).stream()
                .filter(FuzzyVariable::nonNull)
                .collect(Collectors.toList());

        LocalDateTime zeitstempel = ersterDatensatz.getZeitstempel();
        veroeffentlicheKonklussionen(objekt, zeitstempel, konklusionen);
    }

    private static Systemobjekt bestimmeObjekt(Datensatz<?, FuzzyVariable> ersterDatensatz) {
        Systemobjekt systemobjekt = ersterDatensatz.getSystemobjekt();
        if (systemobjekt instanceof Umfelddatensensor) {
            Umfelddatensensor ufds = (Umfelddatensensor) systemobjekt;
            if (ufds.getGehoertZu().isPresent()) {
                systemobjekt = ufds.getGehoertZu().get();
            }
        }
        return systemobjekt;
    }

    private Map<String, ?> erzeugeKontext(Systemobjekt systemobjekt) {
        Map<String, Object> kontext = new HashMap<>();
        fuzzyWerteCache.setzeFuzzyWerteInKontextFuer(systemobjekt, kontext);
        if (systemobjekt instanceof Umfelddatenmessstelle) {
            Umfelddatenmessstelle ufdms = (Umfelddatenmessstelle) systemobjekt;
            ufdms.getUmfelddatensensoren().forEach(ufds -> fuzzyWerteCache.setzeFuzzyWerteInKontextFuer(ufds, kontext));
        }
        return kontext;
    }

    private void veroeffentlicheKonklussionen(Systemobjekt systemobjekt, LocalDateTime zeitstempel, Collection<FuzzyVariable> konklusionen) {
        Collection<Datensatz<?, FuzzyVariable>> fuzzyWerte = konklusionen.stream()
                .map(fv -> new Datensatz<>(zeitstempel, systemobjekt, fv))
                .collect(Collectors.toList());
        if (!fuzzyWerte.isEmpty()) {
            fuzzyWerteAbgeleitet.send(fuzzyWerte);
        }
    }

    public final Event<Collection<Datensatz<?, FuzzyVariable>>> fuzzyWerteAbgeleitet() {
        return fuzzyWerteAbgeleitet;
    }

}
