/*
 * Segment 5 Intelligente Analyseverfahren, SW-Einheit 5.2 Straßensubsegmentanalyse
 * Copyright (C) 2007-2018 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.subsegana.subsegana.domain;

import de.bsvrz.iav.fuzzylib.fuzzylib.*;
import de.bsvrz.iav.subsegana.subsegana.domain.shared.*;
import de.bsvrz.iav.subsegana.subsegana.domain.verkehrsdaten.*;
import de.bsvrz.iav.subsegana.subsegana.shared.*;

import java.util.*;
import java.util.stream.*;

import static de.bsvrz.iav.subsegana.subsegana.domain.verkehrsdaten.Verkehrsgroessen.*;

/**
 * Bestimmt den Verkehrszustand (oder Störfallzustand) zu einem Fuzzy-Wert.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class Verkehrszustandsbestimmung {

    private final Event<Collection<Datensatz<?, Stoerfallzustand>>> verkehrszustandBestimmt = new Event<>();

    public void bestimmeVerkehrszustand(Collection<Datensatz<?, FuzzyVariable>> fuzzyWerte) {
        Collection<Datensatz<?, Stoerfallzustand>> verkehrszustand = fuzzyWerte.stream()
                .filter(d -> VERKEHRSZUSTAND.equals(d.getDatum().getName()))
                .map(Verkehrszustandsbestimmung::bestimmeVerkehrszustand)
                .collect(Collectors.toList());
        verkehrszustandBestimmt.send(verkehrszustand);
    }

    private static Datensatz<?, Stoerfallzustand> bestimmeVerkehrszustand(Datensatz<?, FuzzyVariable> fuzzyWert) {
        Stoerfallzustand zustand = bestimmeStoerfallzustand(fuzzyWert);
        return new Datensatz<>(
                fuzzyWert.getZeitstempel(),
                fuzzyWert.getSystemobjekt(),
                zustand
        );
    }

    private static Stoerfallzustand bestimmeStoerfallzustand(Datensatz<?, FuzzyVariable> fuzzyWert) {
        return fuzzyWert.getDatum().getTerme().stream()
                .filter(t -> t.getZugehoerigkeit().nonNull())
                .filter(t -> t.getZugehoerigkeit().getWert() > 0)
                .max(Comparator.comparing(Term::getZugehoerigkeit))
                .map(Verkehrszustandsbestimmung::bestimmeStoerfallzustand)
                .orElse(new Stoerfallzustand(Stoerfallsituation.KEINE_AUSSAGE, Gueteindex.von(1.0)));
    }

    private static Stoerfallzustand bestimmeStoerfallzustand(Term term) {
        Stoerfallsituation situation;
        Gueteindex index = Gueteindex.von(term.getZugehoerigkeit().getWert());
        if (term.getName().toLowerCase().contains("frei")) {
            situation = Stoerfallsituation.FREIER_VERKEHR;
        } else if (term.getName().toLowerCase().contains("dicht")) {
            situation = Stoerfallsituation.DICHTER_VERKEHR;
        } else if (term.getName().toLowerCase().contains("zäh")) {
            situation = Stoerfallsituation.ZAEHFLIESSENDER_VERKEHR;
        } else if (term.getName().toLowerCase().contains("stau")) {
            situation = Stoerfallsituation.STAU;
        } else {
            situation = Stoerfallsituation.KEINE_AUSSAGE;
            index = Gueteindex.von(1.0);
        }
        return new Stoerfallzustand(situation, index);
    }

    public Event<Collection<Datensatz<?, Stoerfallzustand>>> verkehrszustandBestimmt() {
        return verkehrszustandBestimmt;
    }

}
