/*
 * Segment 5 Intelligente Analyseverfahren, SW-Einheit 5.2 Straßensubsegmentanalyse
 * Copyright (C) 2007-2018 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.subsegana.subsegana.portals;

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.iav.fuzzylib.fuzzylib.*;
import de.bsvrz.iav.subsegana.subsegana.domain.shared.*;
import de.bsvrz.iav.subsegana.subsegana.domain.umfelddaten.*;
import de.bsvrz.iav.subsegana.subsegana.domain.verkehrsdaten.*;
import de.bsvrz.sys.funclib.debug.*;
import de.bsvrz.sys.funclib.operatingMessage.*;

import java.text.*;
import java.util.*;
import java.util.regex.*;

import static de.bsvrz.iav.subsegana.subsegana.portals.Konstanten.*;

/**
 * Sendet Regelbasen an den Datenverteiler.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
@SuppressWarnings("ClassFanOutComplexity")
public class RegelbasisSender extends Kommunikation {

    private static final String ASP_PID = "asp.parameterVorgabe";

    private final Debug log = Debug.getLogger();
    private final MessageSender messageSender = MessageSender.getInstance();
    private final ClientSenderInterface sender = new SenderOhneSendesteuerung();

    public RegelbasisSender(ClientDavInterface verbindung) {
        super(verbindung);
    }

    public void anmelden(Collection<Systemobjekt> objekte) {
        Collection<SystemObject> messquerschnitte = konvertiereObjekte(objekte, obj -> obj instanceof Messquerschnitt);
        doAnmelden(messquerschnitte);
        log.config(MessageFormat.format("Als Sender von Regelbasis für {0} Messquerschnitte angemeldet.", messquerschnitte.size()));

        Collection<SystemObject> strassenteilsegmente = konvertiereObjekte(objekte, obj -> obj instanceof Strassenteilsegment);
        doAnmelden(strassenteilsegmente);
        log.config(MessageFormat.format("Als Sender von Regelbasis für {0} Straßenteilsegmente angemeldet.", strassenteilsegmente.size()));

        Collection<SystemObject> umfelddatenmessstellen = konvertiereObjekte(objekte, obj -> obj instanceof Umfelddatenmessstelle);
        doAnmelden(umfelddatenmessstellen);
        log.config(MessageFormat.format("Als Sender von Regelbasis für {0} Umfelddatenmessstellen angemeldet.", umfelddatenmessstellen.size()));
    }


    private void doAnmelden(Collection<SystemObject> objects) {
        DataDescription dd = datenbeschreibung(ATG_REGELBASIS, ASP_PID);
        try {
            getVerbindung().subscribeSender(sender, objects, dd, SenderRole.sender());
        } catch (OneSubscriptionPerSendData e) {
            log.error("Sendeanmeldung fehlgeschlagen: " + e, Arrays.asList(objects, dd));
            throw new IllegalStateException("Sendeanmeldung fehlgeschlagen", e);
        }
    }

    public void sende(Collection<Datensatz<?, Regelbasis>> datensaetze) {
        ResultData[] results = datensaetze.stream()
                .map(this::konvertiereDatensatz)
                .toArray(ResultData[]::new);
        try {
            getVerbindung().sendData(results);
            log.info(MessageFormat.format("{0} Regelbasen gesendet", datensaetze.size()));
        } catch (SendSubscriptionNotConfirmed e) {
            String message = "Parameter für Regelbasen konnten nicht gesendet werden: " + e;
            log.warning(message, e);
            messageSender.sendMessage(MessageType.SYSTEM_DOMAIN, MessageGrade.WARNING, message);
        }
    }

    private ResultData konvertiereDatensatz(Datensatz<?, Regelbasis> datensatz) {
        SystemObject obj = getModell().getObject(datensatz.getSystemobjekt().getPid());
        AttributeGroup atg = getModell().getAttributeGroup(ATG_REGELBASIS);
        long time = konvertiereZeitstempel(datensatz.getZeitstempel());
        Data data = konvertiereDatum(datensatz.getDatum(), getVerbindung().createData(atg));
        return new ResultData(obj, datenbeschreibung(atg.getPid(), ASP_PID), time, data);
    }

    private static Data konvertiereDatum(Regelbasis regelbasis, Data data) {
        data.getItem("Urlasser").getTextValue("Ursache").setText("Setzen von Standardparametern.");
        data.getItem("Urlasser").getTextValue("Veranlasser").setText("Straßensubsegmentanalyse");
        konvertiereRegeln(regelbasis, data);
        return data;
    }

    private static void konvertiereRegeln(Regelbasis regelbasis, Data data) {
        Data.Array array = data.getArray("Regeln");
        array.setLength(regelbasis.getRegeln().size());
        for (int i = 0; i < array.getLength(); i++) {
            Regel regel = regelbasis.getRegeln().get(i);
            Data dRegel = array.getItem(i);

            konvertiereSicherheitsgrad(regel, dRegel);
            konvertierePraemisse(regel, dRegel);
            konvertiereKonklussion(regel, dRegel);
        }
    }

    private static void konvertiereSicherheitsgrad(Regel regel, Data dRegel) {
        dRegel.getScaledValue("Sicherheitsgrad").set(regel.getSicherheitsgrad().getWert());
    }

    private static void konvertierePraemisse(Regel regel, Data dRegel) {
        List<? extends Ausdruck<Zugehoerigkeit>> praemisse = ((Minimum) regel.getPraemisse()).getOperanden();
        Data.Array aPraemisse = dRegel.getArray("Prämisse");
        aPraemisse.setLength(praemisse.size());
        for (int j = 0; j < aPraemisse.getLength(); j++) {
            Ausdruck<?> ausdruck = praemisse.get(j);
            Data aFuzzyVariable = aPraemisse.getItem(j);
            if (ausdruck instanceof TermVariable) {
                konvertiereFuzzyVariable((TermVariable) ausdruck, aFuzzyVariable);
            } else {
                konvertiereFuzzyVariableMitMehrerenTermen((Maximum) ausdruck, aFuzzyVariable);
            }
        }
    }

    private static void konvertiereFuzzyVariable(TermVariable variable, Data aFuzzyVariable) {
        String term = bestimmeTerm(variable);
        int zyklus = bestimmeZyklus(variable);
        aFuzzyVariable.getTextValue("Variable").setText(variable.getFuzzyVariable());
        aFuzzyVariable.getScaledValue("Zyklus").set(zyklus);
        aFuzzyVariable.getTextArray("Term").set(term);
    }

    private static void konvertiereFuzzyVariableMitMehrerenTermen(Maximum operation, Data aFuzzyVariable) {
        List<?> operanden = operation.getOperanden();

        TermVariable variable = (TermVariable) operanden.get(0);
        int zyklus = bestimmeZyklus(variable);
        aFuzzyVariable.getTextValue("Variable").setText(variable.getFuzzyVariable());
        aFuzzyVariable.getScaledValue("Zyklus").set(zyklus);

        String[] terme = operanden.stream()
                .map(o -> (TermVariable) o)
                .map(RegelbasisSender::bestimmeTerm)
                .toArray(String[]::new);
        aFuzzyVariable.getTextArray("Term").set(terme);
    }

    private static void konvertiereKonklussion(Regel regel, Data dRegel) {
        TermVariable konklusion = regel.getKonklusion();
        Data dKonklussion = dRegel.getItem("Konklusion");
        dKonklussion.getTextValue("Variable").setText(konklusion.getFuzzyVariable());
        dKonklussion.getTextValue("Term").setText(konklusion.getTerm());
    }

    private static String bestimmeTerm(TermVariable v) {
        Matcher m = Pattern.compile("(.*\\D+)(\\d+)").matcher(v.getTerm());
        if (m.matches()) {
            return m.group(1);
        } else {
            return v.getTerm();
        }
    }

    private static int bestimmeZyklus(TermVariable v) {
        Matcher m = Pattern.compile("(.*\\D+)(\\d+)").matcher(v.getTerm());
        if (m.matches()) {
            return Integer.parseInt(m.group(2));
        } else {
            return 0;
        }
    }

}
